@@ -41,6 +41,7 @@
 
 #include "dwarf.h"
 #include "libdwarf.h"
+#include "dwarf_local_malloc.h"
 #include "libdwarf_private.h"
 #include "dwarf_base_types.h"
 #include "dwarf_safe_strcpy.h"
@@ -796,7 +797,7 @@ dwarf_dnames_header(Dwarf_Debug dbg,
         return DW_DLV_NO_ENTRY;
     }
     start_section = dbg->de_debug_names.dss_data;
-    curptr = start_section += starting_offset;
+    curptr = start_section + starting_offset;
     end_section = start_section + section_size;
     dn =  (Dwarf_Dnames_Head)_dwarf_get_alloc(dbg,
         DW_DLA_DNAMES_HEAD, 1);
@@ -809,10 +810,10 @@ dwarf_dnames_header(Dwarf_Debug dbg,
     dn->dn_magic = DWARF_DNAMES_MAGIC;
     dn->dn_section_data = start_section;
     dn->dn_section_size = section_size;
-    dn->dn_section_end = start_section + section_size;
+    dn->dn_section_end = end_section;
     dn->dn_dbg = dbg;
     dn->dn_section_offset = starting_offset;
-    dn->dn_indextable_data = starting_offset + start_section;;
+    dn->dn_indextable_data = curptr;
     remaining = dn->dn_section_size - starting_offset;
     res = read_a_name_table_header(dn,
         starting_offset,
@@ -1694,8 +1695,27 @@ isformrefval(Dwarf_Debug dbg,Dwarf_Half form,
     Dwarf_Unsigned *bytesread,
     Dwarf_Error *error)
 {
-    Dwarf_Unsigned localval =0;
+    Dwarf_Unsigned localval = 0;
 
+    if (poolptr >= endpool) {
+        dwarfstring m;
+        const char * formname = 0;
+
+        dwarfstring_constructor(&m);
+        dwarfstring_append_printf_u(&m,
+            "DW_DLE_DEBUG_NAMES_OFF_END: "
+            "Reading data of form 0x%02x ",form);
+        dwarf_get_FORM_name((unsigned int)form,&formname);
+        dwarfstring_append_printf_s(&m,
+            "%s from entrypool would read "
+            "off the end of the pool",
+            (char *)formname);
+        _dwarf_error_string(dbg,error,
+            DW_DLE_DEBUG_NAMES_OFF_END,
+            dwarfstring_string(&m));
+        dwarfstring_destructor(&m);
+        return DW_DLV_ERROR;
+    }
     switch(form) {
     case DW_FORM_ref1:
         *val = *poolptr;
@@ -1771,6 +1791,7 @@ int dwarf_dnames_entrypool_values(Dwarf_Dnames_Head dn,
         /* make error or harmless error? */
         return DW_DLV_NO_ENTRY;
     }
+
     if (index_of_abbrev >= dn->dn_abbrev_instance_count) {
         /* make error or harmless error? */
         return DW_DLV_NO_ENTRY;
@@ -1854,6 +1875,10 @@ int dwarf_dnames_entrypool_values(Dwarf_Dnames_Head dn,
 
             array_of_offsets[n] = val;
             continue;
+        } else if (form == DW_FORM_flag_present) {
+            array_of_offsets[n] = 1;
+            /* No change to poolptr or pooloffset */
+            continue;
         } else {
             Dwarf_Unsigned val = 0;
             res =isformrefval(dbg,form,poolptr,
@@ -1863,7 +1888,7 @@ int dwarf_dnames_entrypool_values(Dwarf_Dnames_Head dn,
             }
             if (res == DW_DLV_OK) {
                 poolptr += bytesread;
-                if (poolptr > endpool) {
+                if (poolptr >= endpool) {
                     _dwarf_error_string(dbg,error,
                         DW_DLE_DEBUG_NAMES_ENTRYPOOL_OFFSET,
                         "DW_DLE_DEBUG_NAMES_ENTRYPOOL_OFFSET:"
@@ -1871,7 +1896,6 @@ int dwarf_dnames_entrypool_values(Dwarf_Dnames_Head dn,
                         " of the entrypool");
                     return DW_DLV_ERROR;
                 }
-                poolptr += bytesread;
                 pooloffset += bytesread;
                 array_of_offsets[n] = val;
                 continue;
